"""Add enable_parallel_execution to Tools table

Revision ID: e67961ed7c32
Revises: 8149a781ac1b
Create Date: 2025-10-17 15:47:00.447066

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e67961ed7c32"
down_revision: Union[str, None] = "8149a781ac1b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("tools", sa.Column("enable_parallel_execution", sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("tools", "enable_parallel_execution")
    # ### end Alembic commands ###
