"""npm requirements in tools

Revision ID: d007f4ca66bf
Revises: 74e860718e0d
Create Date: 2025-08-04 13:40:32.707036

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d007f4ca66bf"
down_revision: Union[str, None] = "74e860718e0d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("tools", sa.Column("npm_requirements", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("tools", "npm_requirements")
    # ### end Alembic commands ###
