"""Add tools_used field to run_metrics table

Revision ID: 6756d04c3ddb
Revises: e67961ed7c32
Create Date: 2025-10-17 14:52:53.601368

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "6756d04c3ddb"
down_revision: Union[str, None] = "e67961ed7c32"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("run_metrics", sa.Column("tools_used", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("run_metrics", "tools_used")
    # ### end Alembic commands ###
